﻿using System.Drawing;
using System.Linq;
using SolidFramework.Model.Plumbing;

namespace PdfToHtml {
    static class HelperMethods {

        public static string ToHexString(this Color color, bool exists=true) {
            if (!exists) {
                return null;
            }
            return "#" + color.R.ToString("X2") + color.G.ToString("X2") + color.B.ToString("X2");
        }

        public static int GetZIndex(this Shape shape, int defaultValue) {
            var wrappingType = shape.WrappingType;
            if (wrappingType == WrappingType.Behind) {
                return -10;
            } else if (wrappingType == WrappingType.Infront) {
                return 10;
            }
            return defaultValue;
        }
        
        public static double GetBorderSize(this LineStyle border) {
            if (border == null) {
                return 0;
            }
            double size = border.Weight;
            switch (border.CompoundType) {
                case LineCompoundType.ThickThin:
                case LineCompoundType.ThinThick:
                    return size * 1.5;
                case LineCompoundType.ThinThickThin:
                    return size * 2;
                case LineCompoundType.Double:
                    return size * 3;
                case LineCompoundType.Triple:
                    return size * 5;
                default:
                    return size;
            }
        }
        
        public static PointF Copy(this PointF point) {
            return new PointF(point.X, point.Y);
        }
        public static PointF OffsetBy(this PointF point, double x, double y) {
            var result = point.Copy();
            result.X += (float)x;
            result.Y += (float)y;
            return result;
        }
        public static PointF OffsetBy(this PointF point, PointF offset) {
            return point.OffsetBy(offset.X, offset.Y);
        }
        
        private static string GetAlphaList(int number) {
            return new string((char) ('A' + (number % 26)), 1 + (number / 26));
        }

        private static string GetRomanList(int number) {
            var romanNumerals = new[] {
                new[] { "", "I", "II", "III", "IV", "V", "VI", "VII", "VIII", "IX" }, // ones
                new[] { "", "X", "XX", "XXX", "XL", "L", "LX", "LXX", "LXXX", "XC" }, // tens
                new[] { "", "C", "CC", "CCC", "CD", "D", "DC", "DCC", "DCCC", "CM" }, // hundreds
                new[] { "", "M", "MM", "MMM" } // thousands
            };
            if (number >= 4000) {
                return number.ToString(); // numbers that big require unicode characters
            }
            var digits = number.ToString().Reverse().ToArray();
            string romanNumeral = "";
            int i = digits.Length;
            while (i-- > 0) {
                romanNumeral += romanNumerals[i][int.Parse(digits[i].ToString())];
            }
            return romanNumeral;
        }

        public static string FormatPageNumber(this PageNumberFormat format, int number) {
            switch (format) {
                case PageNumberFormat.None:
                    return "";
                case PageNumberFormat.Decimal:
                    return number.ToString();
                case PageNumberFormat.LowerLatin:
                    return GetAlphaList(number).ToLower();
                case PageNumberFormat.UpperLatin:
                    return GetAlphaList(number);
                case PageNumberFormat.LowerRoman:
                    return GetRomanList(number).ToLower();
                case PageNumberFormat.UpperRoman:
                    return GetRomanList(number);
                default:
                    return number.ToString();
            }
        }
    }
}