﻿using SolidFramework.Converters.Plumbing;
using SolidFramework.Model;
using SolidFramework.Model.Plumbing;
using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using SolidFramework.Model.Layout;
using Section = SolidFramework.Model.Plumbing.Section;

namespace PdfToHtml.Models {

    class FileModel: ParentModel, IDisposable {
        public double Scale { get; set; }
        public string PdfPath { get; private set; }
        public CoreModel Model { get; private set; }
        public bool ExactMode { get; private set; }
        public bool AddNavigation { get; set; }
        public bool EmbedImages { get; set; }

        public List<PageModel> Pages { get; private set; }

        public bool UseRelativeScale { get; set; }
        public double RemSize { get; private set; }
        
        private Dictionary<int, LayoutObject> layoutObjects;
        private Dictionary<LayoutObject, int> layoutPageNumbers;
        private int layoutPages;

        private void WalkLayout(IEnumerable<LayoutObject> collection, int pageNumber) {
            int page = pageNumber;
            foreach (var obj in collection) {
                if (pageNumber < 1) {
                    if (obj is LayoutPage) {
                        layoutPages += 1;
                    }
                    page += 1;
                }
                int reliablePage = obj.GetPageNumber();
                if (reliablePage > 0) {
                    page = reliablePage;
                }
                int id = obj.GetID();
                if (id > 0) {
                    if (!layoutObjects.ContainsKey(id)) {
                        layoutObjects.Add(id, obj);
                    }
                }
                layoutPageNumbers.Add(obj, page);
                var objCollection = obj as IEnumerable<LayoutObject>;
                if (objCollection != null) {
                    WalkLayout(objCollection, page);
                }
            }
        }

        public LayoutObject FindLayoutObject(int id) {
            if (id >= 0 && layoutObjects != null && layoutObjects.ContainsKey(id)) {
                return layoutObjects[id];
            }
            return null;
        }

        private LayoutObject FindFirstLayoutObjectFor(SolidObject obj) {
            var layout = FindLayoutObject(obj == null? -1 : obj.GetID());
            if (layout != null) {
                return layout;
            }
            var objects = obj as IEnumerable<SolidObject>;
            if (objects != null) {
                foreach (var child in objects) {
                    layout = FindFirstLayoutObjectFor(child);
                    if (layout != null) {
                        return layout;
                    }
                }
            }
            return null;
        }

        public int GetPageNumberFor(SolidObject obj, int defaultValue=-1) {
            var layout = FindFirstLayoutObjectFor(obj);
            if (layout != null) {
                if (layoutPageNumbers.ContainsKey(layout)) {
                    return layoutPageNumbers[layout];
                }
            }
            return defaultValue;
        }

        public FileModel(string pdfPath, bool exactMode, bool useRelativeScale=true, bool addNavigation=true): base(null) {
            Scale = 1;
            RemSize = 10;
            ExactMode = exactMode;
            UseRelativeScale = useRelativeScale;
            AddNavigation = addNavigation;
            EmbedImages = true;
            var options = new PdfOptions();
            options.ConvertMode = ConvertMode.Document;
            options.ExposeTargetDocumentPagination = true;
            options.ExposeSourceDocumentPagination = true;
            options.DetectLanguage = true;
            options.TransformPrivateUnicode = false;
            options.DetectToc = true;
            options.MarkupAnnotConversionType = MarkupAnnotConversionType.Never;
            //options.MergeParagraphIndents = false;
            //options.AverageCharacterScaling = false;
            ZIndex = 0;
            PdfPath = pdfPath;
            Model = CoreModel.Create(pdfPath, options);
            Pages = new List<PageModel>();
            var layoutDocument = Model.GetLayout();
            if (layoutDocument == null) {
                throw new InvalidOperationException("PdfToHtml requires the layout model");
            }
            layoutPages = 0;
            layoutObjects = new Dictionary<int, LayoutObject>();
            layoutPageNumbers = new Dictionary<LayoutObject, int>();
            WalkLayout(layoutDocument, 0);

            // walk the pages
            var defaultSection = Model.DefaultSection;
            bool evenOdd = defaultSection.HasDifferentHeaderFooterForEvenPages;
            var info = new HeaderFooterInfo(null, 0, evenOdd, null);
            bool evenPage = false;
            PageModel currentPage = null;
            List<TableOfContentsModel> unfinishedTOCs = new List<TableOfContentsModel>();
            foreach (var obj in Model.Topic) {
                var section = obj as Section;
                if (section == null) {
                    if (obj is HeaderFooterBlockCollection) {
                        info = new HeaderFooterInfo(obj as HeaderFooterBlockCollection, Pages.Count, evenOdd, info);
                        evenPage = info.StartsEven;
                    }
                    continue;
                }
                // A Document mode Section could span multiple LayoutPages or could share its LayoutPage with other Sections
                foreach (var child in section) {
                    if (currentPage == null) {
                        currentPage = AddPage(section, info, evenOdd, ref evenPage, unfinishedTOCs);
                    }
                    int layoutPageNumber = GetPageNumberFor(child);
                    if (layoutPageNumber > 0) {
                        if ((layoutPageNumber != currentPage.Index)) {
                            while (layoutPageNumber > currentPage.Index) {
                                currentPage.FinishedLoadingContents();
                                currentPage = AddPage(section, info, evenOdd, ref evenPage, unfinishedTOCs);
                            }
                        }
                    }
                    if (child is TOC) {
                        currentPage.TOC = new TableOfContentsModel(currentPage, (TOC)child);
                        currentPage.Contents.Add(currentPage.TOC);
                        if (currentPage.TOC.AddItemsToLayoutPage(currentPage)) {
                            currentPage.TOC.FinishedLoadingContents();
                        } else {
                            unfinishedTOCs.Add(currentPage.TOC);
                        }
                        continue;
                    }
                    currentPage.AddContent(child);
                }
            }
            if (currentPage != null) {
                currentPage.FinishedLoadingContents();
            }
            while (Pages.Count < layoutPages) {
                AddPage(defaultSection, info, evenOdd, ref evenPage, unfinishedTOCs);
            }
            for (int i = unfinishedTOCs.Count - 1; i >= 0; i--) {
                unfinishedTOCs[i].FinishedLoadingContents();
            }
            ParagraphStyles = new Dictionary<string, Dictionary<string, object>>();
        }

        public void SaveHtml(string htmlPath, string resourcePath, string resourceUrl) {
            Directory.CreateDirectory(Path.GetDirectoryName(htmlPath));
            var html = new HtmlBuilder(resourcePath, resourceUrl);
            SaveHtml(html, 0, 0);
            System.IO.File.WriteAllText(htmlPath, html.ToString(), Encoding.UTF8);
        }

        public Dictionary<string, Dictionary<string, object>> ParagraphStyles { get; private set; }

        public override void SaveHtml(HtmlBuilder html, double x, double y) {
            int maxWidth = 0, maxHeight = 0;
            foreach (var page in Pages) {
                maxWidth = Math.Max(maxWidth, (int)page.Width);
                maxHeight = Math.Max(maxHeight, (int)page.Height);
            }
            if (maxWidth < 1) {
                maxWidth = 612;
            }
            if (maxHeight < 1) {
                maxHeight = 792;
            }
            ParagraphStyles.Clear();
            html.Raw("<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">");
            html.Open("html", style: new Dictionary<string, object> { { "font-size", RemSize.ToString("0.0") + "pt" } });
            html.Open("head");
            html.Inline("title", content: Path.GetFileNameWithoutExtension(PdfPath));
            html.Open("style");
            html.Raw("p { margin: 0; }");
            html.Raw(".page { margin-top: 10pt; margin-bottom: 10pt; position: relative; }");
            html.Raw("span.position, p.paragraph { position: absolute; display: block; }");
            html.Raw("span.position { transform-origin: left; text-align: left; white-space: nowrap; }");
            html.Raw("span.style { white-space: nowrap; }");
            html.Raw("td { padding: 0px; }");
            html.Raw("a.link { color: inherit; text-decoration: inherit; }");
            html.Raw("div.group, div.textbox, svg.graphic, img.image, div.rotation { position: absolute; }");
            html.MultiLineRaw(@"
.noselect {
    -webkit-touch-callout: none;
    -webkit-user-select: none;
    -khtml-user-select: none;
    -moz-user-select: none;
    -ms-user-select: none;
    user-select: none;
}".TrimStart());
            if (AddNavigation && UseRelativeScale) {
                html.Raw("body { overflow: scroll; }");
                html.MultiLineRaw(@"
#navigation-pane .control {
    display: block;
    width: 20px;
    height: 20px;
    margin: 2px;
    font-size: 14px;
    text-align: center;
    line-height: 20px;
    border: 1px solid blue;
    border-radius: 12px;
    background: rgba(230, 230, 255, 0.4);
    cursor: pointer;
    color: blue;
}
#navigation-pane .control:hover {
    background: rgba(120, 120, 255, 0.4);
}
@media print { #navigation-pane { display: none; } }
@media print { html, body { font-size: " + RemSize + @"pt !important; } }
".TrimStart());
            }
            html.Raw("@media print { @page { size: " + maxWidth.ToString("0.0") + "pt " + maxHeight.ToString("0.0") + "pt; margin: 0pt 0pt 0pt 0pt; } html, body { width: " + maxWidth.ToString("0.0") + "pt; height: " + maxHeight.ToString("0.0") + "pt; font-size: 10pt; } }");
            html.Raw("@media print { .page { margin-top: 0pt; margin-bottom: 0pt; } }");
            foreach (var template in Model.StyleTemplates) {
                var s = template as ParagraphStyleTemplate;
                if (s == null) {
                    continue;
                }
                var style = s.Style;
                string name = s.Name.ToLower().Replace(" ", "-");
                string css = "p." + name + " { ";
                var styles = new Dictionary<string, object>();
                string fontname = (style.Font != null ? style.Font.Name : "arial");
                if (fontname.Contains(" ")) {
                    fontname = "'" + fontname + "'";
                }
                styles["z-index"] = "0";
                styles["font-family"] = fontname + ", serif";
                css += "font-family: " + fontname + ", serif; ";
                double fontSize = style.FontSize;
                if (fontSize < 0.25) {
                    fontSize = Math.Max(6, -fontSize);
                }
                styles["font-size"] = Pt(fontSize);
                css += "font-size: " + Pt(fontSize) + "; ";
                if (style.HasTextColor) {
                    styles["color"] = style.TextColor.ToHexString();
                    css += "color: " + styles["color"] + "; ";
                }
                if (style.FontWeight != 0) {
                    styles["font-weight"] = style.FontWeight.ToString();
                    css += "font-weight: " + styles["font-weight"] + "; ";
                }
                html.Raw(css + "}");
                ParagraphStyles[name] = styles;
            }
            html.Close("style");
            if (AddNavigation && UseRelativeScale) {
                html.Open("script");
                html.MultiLineRaw(@"
var defaultZoom = " + RemSize + @";
var pageWidth = " + maxWidth + @";
var pageHeight = " + maxHeight + @";

function getFontSize(style) {
    style = style || document.documentElement.style;
    var fontSize = parseFloat(style.getPropertyValue('font-size').replace('pt', '') || 'NaN');
    if (isNaN(fontSize)) { fontSize = defaultZoom; }
    return fontSize;
}
function setFontSize(style, fontSize) {
    var body = document.body;
    var vScrollPercent = (body.scrollTop/* + body.clientHeight / 2*/) / Math.max(body.scrollHeight, 1);
    var hScrollPercent = (body.scrollLeft + body.clientWidth / 2) / Math.max(body.scrollWidth, 1);
    style.setProperty('font-size', fontSize + 'pt');
    body.scrollTop = (vScrollPercent * Math.max(body.scrollHeight, 0))/* - body.clientHeight / 2*/;
    body.scrollLeft = hScrollPercent * Math.max(body.scrollWidth, 0) - body.clientWidth / 2;
    var zoomLevelButton = document.getElementById('zoom-level-button');
    var zoomLevel = (100 * fontSize / defaultZoom).toPrecision(3);
    if (zoomLevel.indexOf('.0', zoomLevel.length - 2) !== -1) {
        zoomLevel = zoomLevel.replace('.0', '');
    }
    zoomLevelButton.innerHTML = zoomLevel;
}

function nextZoomLevel(currentZoom, zoomOut) {
    var zoomLevels = [0.25, 0.333, 0.5, 0.667, 0.75, 0.8, 0.9, 1, 1.1, 1.25, 1.5, 1.75, 2, 2.5, 3, 4, 5];
    var currentLevel = currentZoom / defaultZoom;
    var i, zoomLevelsLength = zoomLevels.length, nextZoomLevel = 1;
    if (zoomOut) {
        var nextZoomLevel = zoomLevels[0];
        for (i = 0; i < zoomLevelsLength; i++) {
            if (currentLevel < zoomLevels[i] + 0.01) { break; }
            nextZoomLevel = zoomLevels[i];
        }
        return defaultZoom * nextZoomLevel;
    }
    nextZoomLevel = zoomLevels[zoomLevelsLength - 1];
    for (i = zoomLevelsLength - 1; i >= 0; i--) {
        if (currentLevel > zoomLevels[i] - 0.01) { break; }
        nextZoomLevel = zoomLevels[i];
    }
    return defaultZoom * nextZoomLevel;
}

function zoomIn() {
    var html = document.documentElement;
    var style = html.style;
    var fontSize = getFontSize(style);
    setFontSize(style, nextZoomLevel(fontSize, false));
}
function zoomOut() {
    var html = document.documentElement;
    var style = html.style;
    var fontSize = getFontSize(style);
    setFontSize(style, nextZoomLevel(fontSize, true));
}
function resetZoom() {
    var html = document.documentElement;
    var style = html.style;
    var fontSize = getFontSize(style);
    setFontSize(style, defaultZoom);
}
function fillWidth() {
    var html = document.documentElement;
    setFontSize(html.style, defaultZoom * ((window.innerWidth - 40) * 0.75) / pageWidth);
}
function fitPage() {
    var html = document.documentElement;
    var widthScale = defaultZoom * ((window.innerWidth - 40) * 0.75) / pageWidth;
    var heightScale = defaultZoom * ((window.innerHeight - 40) * 0.75) / pageHeight
    setFontSize(html.style, Math.min(widthScale, heightScale));
}

function addClickHandler(element, handler) {
    if (element.addEventListener) {
        element.addEventListener('click', handler, false);
    } else {
        element.attachEvent('onclick', handler);
    }
}

window.onload = function() {
    addClickHandler(document.getElementById('zoom-in-button'), zoomIn);
    addClickHandler(document.getElementById('zoom-out-button'), zoomOut);
    addClickHandler(document.getElementById('zoom-level-button'), resetZoom);
    addClickHandler(document.getElementById('fill-width-button'), fillWidth);
    addClickHandler(document.getElementById('fit-page-button'), fitPage);
}
".TrimStart(), "");
                html.Close("script");
            }
            html.Close("head");
            html.Open("body", null, new Dictionary<string, object> { { "background", "gainsboro"}, { "margin", "0" } });
            if (AddNavigation && UseRelativeScale) {
                html.Open("div", "noselect", style: new Dictionary<string, object> {
                    { "position", "fixed" },
                    { "right", "5px" },
                    { "bottom", "5px" },
                    { "height", "128px" },
                    { "width", "24px" },
                    { "z-index", "500" }
                }, attributes: new Dictionary<string, object> { { "id", "navigation-pane" } });
                html.Inline("span", "control fill-width", "Fill", new Dictionary<string, object> {
                    {"font-size", "10px" }
                }, attributes: new Dictionary<string, object> { { "id", "fill-width-button" } });
                html.Inline("span", "control fit-page", "Fit", new Dictionary<string, object> {
                    {"font-size", "10px" }, { "margin-bottom", "10px" }
                }, attributes: new Dictionary<string, object> { { "id", "fit-page-button" } });
                html.Inline("span", "control zoom-in", "+", attributes: new Dictionary<string, object> { { "id", "zoom-in-button" } });
                html.Inline("span", "control zoom-level", "100", new Dictionary<string, object> {
                        {"font-size", "10px" }, { "overflow-x", "hidden" }
                }, attributes: new Dictionary<string, object> { { "id", "zoom-level-button" } });
                html.Inline("span", "control zoom-out", "-", attributes: new Dictionary<string, object> { { "id", "zoom-out-button" } });
                html.Close("div");
            }
            html.Open("div", null, new Dictionary<string, object> {
                { "position", "relative" },
                { "width", Pt(maxWidth) },
                { "margin", "auto" }
            });
            foreach (var page in Pages) {
                page.SaveHtml(html, x, y);
            }
            html.Close("div");
            html.Close("body");
            html.Close("html");
        }

        private PageModel AddPage(Section section, HeaderFooterInfo info, bool evenOdd, ref bool evenPage, List<TableOfContentsModel> unfinishedTOCs) {
            var page = new PageModel(this, section, info, CalculatePageType(evenOdd, evenPage), unfinishedTOCs);
            evenPage = evenOdd && !evenPage;
            Pages.Add(page);
            Contents.Add(page);
            return page;
        }

        private static HeaderFooterPageType CalculatePageType(bool evenOdd, bool evenPage) {
            if (!evenOdd) {
                return HeaderFooterPageType.Both;
            }
            return evenPage? HeaderFooterPageType.Even : HeaderFooterPageType.Odd;
        }

        public void Dispose() {
            if (Model != null) {
                Model.Close();
                Model.Dispose();
                Model = null;
            }
        }
    }
}
