﻿using System;
using System.Collections.Generic;
using SolidFramework.Model.Plumbing;

namespace PdfToHtml.Models {

    class HeaderFooterInfo {
        public PageModel FirstPageUsedOn;
        public int PageNumberOffset { get; private set; } // number to offset the actual page number by
        public PageNumberFormat PageNumberFormat { get; private set; } // format used to display the page number
        public bool StartsEven { get; private set; } // whether the first page to use this will be even
        public bool IsEvenOdd { get; private set; } // whether the document has even & odd headers/footers

        private HeaderFooter[] headers = new HeaderFooter[2]; // holds the odd(0) and even(1) headers
        private HeaderFooter[] footers = new HeaderFooter[2]; // holds the odd(0) and even(1) footers

        public HeaderFooter GetHeader(bool evenPage) {
            return headers[IsEvenOdd && evenPage? 1 : 0];
        }
        public HeaderFooter GetFooter(bool evenPage) {
            return footers[IsEvenOdd && evenPage? 1 : 0];
        }


        public HeaderFooterInfo(HeaderFooterBlockCollection hfbc, int pageCount, bool isEvenOdd, HeaderFooterInfo previous) {
            IsEvenOdd = isEvenOdd;
            FirstPageUsedOn = null;
            if (hfbc == null) {
                PageNumberOffset = pageCount;
                PageNumberFormat = PageNumberFormat.None;
                return;
            }
            if (previous != null) {
                headers[0] = previous.headers[0];
                headers[1] = previous.headers[1];
                footers[0] = previous.footers[0];
                footers[1] = previous.footers[1];
            }
            int pageNumber = hfbc.PageNumberStartValue;
            PageNumberOffset = pageNumber - pageCount;
            PageNumberFormat = hfbc.PageNumberFormat;
            StartsEven = IsEvenOdd && pageNumber % 2 == 0;
            bool isEvenHeader = StartsEven, isEvenFooter = StartsEven;
            foreach (var obj in hfbc) {
                var hf = obj as HeaderFooter;
                if (hf == null) {
                    throw new NotSupportedException("HeaderFooterBlockCollection is expected to only hold HeaderFooters!");
                }
                if (hf.IsHeader) {
                    headers[(isEvenHeader ? 1 : 0)] = hf;
                    isEvenHeader = !isEvenHeader;
                } else {
                    footers[(isEvenFooter ? 1 : 0)] = hf;
                    isEvenFooter = !isEvenFooter;
                }
            }
        }
    }
    
    public enum HeaderFooterPageType {
        Both,
        Even,
        Odd
    }

    class HeaderFooterModel : ParentModel {
        public HeaderFooter Model { get; private set; }

        public bool IsHeader { get; private set; }
        
        public override void SaveHtml(HtmlBuilder html, double x, double y) {
            html.Open("div", IsHeader? "header" : "footer", new Dictionary<string, object> {
                { "width", Pt(Width) },
                { "height", Pt(Height) },
                { "position", "absolute" },
                { "display", "block" },
                { "z-index", ZIndex },
                { "top", Pt(Top-y) },
                { "left", Pt(Left-x) }
            });
            foreach (var child in Contents) {
                child.SaveHtml(html, Left, Top);
            }
            html.Close("div");
        }

        public HeaderFooterModel(PageModel page, HeaderFooter headerFooter): base(page) {
            ZIndex = Parent.ZIndex;
            Model = headerFooter;
            IsHeader = Model.IsHeader;

            var contents = Model as IEnumerable<SolidObject>;
            if (contents != null) {
                foreach (var obj in contents) {
                    AddContent(obj);
                }
            }
        }

        public bool IsEmpty() {
            foreach (var obj in Contents) {
                var paragraph = obj as ParagraphModel;
                if (paragraph != null) {
                    if (string.IsNullOrWhiteSpace(paragraph.AllText)) {
                        continue; 
                    }
                }
                return false;
            }
            return true;
        }
    }
}