﻿using SolidFramework.Model.Plumbing;
using System.Collections.Generic;
using System.Drawing;

namespace PdfToHtml.Models {

    sealed class PageModel: ParentModel {
        public Section Model { get; private set; }
        public HeaderFooterInfo HeaderFooterInfo { get; private set; }
        public HeaderFooterPageType HeaderFooterPageType { get; private set; }

        public TableOfContentsModel TOC { get; set; }
        public List<TableOfContentsModel> ExtraneousTOCs { get; private set; }

        public PageModel(FileModel file, Section section, HeaderFooterInfo headerFooterInfo, HeaderFooterPageType headerFooterPageType, List<TableOfContentsModel> unfinishedTOCs) : base(file) {
            ZIndex = Parent.ZIndex;
            Model = section;
            ExtraneousTOCs = new List<TableOfContentsModel>();
            HeaderFooterPageType = headerFooterPageType;
            HeaderFooterInfo = headerFooterInfo;
            if (HeaderFooterInfo.FirstPageUsedOn == null) {
                HeaderFooterInfo.FirstPageUsedOn = this;
            }
            Width = Model.PageWidth;
            Height = Model.PageHeight;
            Top = 0;
            Left = 0;
            if (Index > 1) {
                Top = File.Pages[Index-2].Top + File.Pages[Index-2].Height;
            }
            
            var header = HeaderFooterInfo.GetHeader(HeaderFooterPageType == HeaderFooterPageType.Even);
            if (header != null) {
                var headerModel = new HeaderFooterModel(this, header);
                if (!headerModel.IsEmpty()) {
                    Contents.Add(headerModel);
                }
            }

            if (unfinishedTOCs.Count > 0) {
                for (int i = unfinishedTOCs.Count - 1; i >= 0; i--) {
                    var toc = unfinishedTOCs[i];
                    if (toc.AddItemsToLayoutPage(this)) {
                        toc.FinishedLoadingContents();
                        unfinishedTOCs.RemoveAt(i);
                    }
                    Contents.Add(toc);
                }
            }
        }

        public override void SaveHtml(HtmlBuilder html, double x, double y) {
            html.Open("div", "page", new Dictionary<string, object> {
                { "width", Pt(Width) },
                { "height", Pt(Height) },
                { "z-index", ZIndex },
                { "background", (Model.HasPageBackgroundColor ? Model.PageBackgroundColor : Color.White).ToHexString() }
            });
            foreach (var toc in ExtraneousTOCs) {
                toc.SaveHtml(this, html, 0, 0);
            }
            foreach (var child in Contents) {
                if (child is TableOfContentsModel) {
                    if (child.Page == this) {
                        child.SaveHtml(html, 0, 0);
                    }
                    continue;
                }
                child.SaveHtml(html, 0, 0);
            }
            html.Close("div");
        }

        public void FinishedLoadingContents() {
            var footer = HeaderFooterInfo.GetFooter(HeaderFooterPageType == HeaderFooterPageType.Even);
            if (footer != null) {
                var footerModel = new HeaderFooterModel(this, footer);
                if (!footerModel.IsEmpty()) {
                    Contents.Add(footerModel);
                }
            }
        }
    }
}
