﻿using System;
using System.Collections.Generic;
using System.Drawing;
using SolidFramework.Model.Layout;
using SolidFramework.Model.Plumbing;

namespace PdfToHtml.Models {
    class ParagraphModel: BaseModel {
        public Paragraph Model { get; private set; }

        public string AllText { get; set; }

        public ParagraphModel(ParentModel parent, Paragraph paragraph) : base(parent) {
            ZIndex = Parent.ZIndex;
            Model = paragraph;
            if (!GetLayoutInfo(Model)) {
                if (Parent.Contents.Count > 0) {
                    var previous = Parent.Contents[Parent.Contents.Count-1];
                    Left = previous.Left;
                    Top = previous.Top + previous.Height;
                    Width = previous.Width;
                    Height = Math.Min(Math.Max(previous.Height, 10), 50);
                } else {
                    Left = Parent.Left;
                    Top = Parent.Top;
                    Width = Parent.Width;
                    Height = Math.Min(Math.Max(8, Parent.Height), 20);
                }
            }

            AllText = "";
            foreach (var run in Model) {
                string text = run.Text;
                if (run.IsPageNumberField) {
                    int pageNumber = Page.Index + Page.HeaderFooterInfo.PageNumberOffset - 1;
                    text = Page.HeaderFooterInfo.PageNumberFormat.FormatPageNumber(pageNumber);
                }
                AllText += text;
            }
        }

        public override void SaveHtml(HtmlBuilder html, double x, double y) {
            if (AllText.Length < 1) {
                return;
            }
            var paragraphStyle = Model.Style;
            var alignment = paragraphStyle.Alignment;
            double top = Top, left = Left;
            if ((TextRotatation > 85) && (TextRotatation < 95)) {
                left = x + (Top-y);
                top = y - (Left-x);
            } else if ((TextRotatation > 265) && (TextRotatation < 275)) {
                left = x + (Top-y);
                top = y + (Left-x);
            }
            var s = new Dictionary<string, object> {
                { "width", Pt(Width) },
                { "height", Pt(Height) },
                { "z-index", ZIndex },
                { "top", Pt(top-y) },
                { "left", Pt(left-x) },
                //{ "background", Colors.AntiqueWhite.ToHexString(true) },
                { "text-align", (alignment == Alignment.Center || alignment == Alignment.Justified)? "center" : (alignment == Alignment.Right? "right" : "left") }
            };
            GetParagraphStyles(s);
            string className = "paragraph";
            string templateName = null;
            if (paragraphStyle.HasHighlightColor) {
                s["background"] = paragraphStyle.HighlightColor.ToHexString();
            }
            var template = paragraphStyle.BasedOn as ParagraphStyleTemplate;
            Dictionary<string, object> templateStyle = null;
            if (template != null) {
                templateName = template.Name.ToLower().Replace(" ", "-");
            }
            if (templateName == null) {
                templateName = "body-text";
            }
            if (File.ParagraphStyles.ContainsKey(templateName)) {
                templateStyle = File.ParagraphStyles[templateName];
                bool matches = template != null;
                foreach (string key in templateStyle.Keys) {
                    if (s.ContainsKey(key) && s[key].ToString() == templateStyle[key].ToString()) {
                        matches = true;
                        s.Remove(key);
                    }
                }
                if (matches) {
                    className += " " + templateName;
                } else {
                    templateStyle = null;
                }
            }
            html.Open("p", className, s);
            if (templateStyle != null) {
                foreach (string key in templateStyle.Keys) {
                    if (!s.ContainsKey(key)) {
                        s[key] = templateStyle[key];
                    }
                }
            }
            if (!File.ExactMode || !AddRunWords(html, Left, Top, s)) {
                // a fallback that should only happen when a paragraph has no layout info
                Hyperlink hyperlink = null;
                foreach (var run in Model) {
                    if (run.Hyperlink != hyperlink) {
                        DoHyperlink(html, run.Hyperlink, hyperlink);
                        hyperlink = run.Hyperlink;
                    }
                    var style = run.Style;
                    double fontsize = style.FontSize;
                    if (fontsize < 0) {
                        fontsize = Math.Max(6, -fontsize);
                    }
                    string fontname = (style.Font != null ? style.Font.Name : "arial");
                    if (fontname.Contains(" ")) {
                        fontname = "'" + fontname + "'";
                    }
                    Dictionary<string, object> a = null;
                    if (run.BookmarkStart != null) {
                        a = new Dictionary<string, object> { { "id", GetBookmarkId(run.BookmarkStart) } };
                    }
                    html.Inline("span", "run", run.Text.Replace(char.ToString((char)14), ""), new Dictionary<string, object> {
                        {"font-family", fontname + ", serif"},
                        {"font-size", Pt(fontsize)},
                        {"color", style.HasTextColor ? style.TextColor.ToHexString() : "inherit"},
                        {"font-weight", style.FontWeight},
                        {"text-decoration", style.UnderlineType == SolidFramework.Model.Plumbing.BorderLineType.None ? "none" : "underline"},
                    }, a);
                }
                DoHyperlink(html, null, hyperlink);
            }
            html.Close("p");
        }

        private void DoHyperlink(HtmlBuilder html, Hyperlink hyperlink, Hyperlink old) {
            if (old != null) {
                html.Close("a");
            }
            if (hyperlink != null) {
                string address = "javascript:void(0)";
                if (hyperlink is InternalHyperlink) {
                    address = "#" + GetBookmarkId(((InternalHyperlink) hyperlink).Bookmark);
                } else if (hyperlink is ExternalHyperlink) {
                    address = ((ExternalHyperlink) hyperlink).Url.ToString();
                }
                html.Open("a", "link", attributes: new Dictionary<string, object> {
                    {"href", address}
                });
            }
        }

        private string GetBookmarkId(Bookmark bookmark) {
            return bookmark.Name.ToLower().Replace(" ", "-");
        }

        private bool AddRunWords(HtmlBuilder html, double x, double y, Dictionary<string, object> paragraphStyle) {
            var runWords = new RunWords();
            var layout = File.FindLayoutObject(Model.GetID()) as LayoutParagraph;
            if (layout == null) {
                return false;
            }
            
            bool hasChunkLines = false;
            var textLines = layout.ChunkLines;
            if (textLines != null) {
                foreach (var textLine in textLines) {
                    foreach (var textChunk in textLine) {
                        string text = textChunk.Text;
                        var bounds = textChunk.Bounds;
                        if ((text == null) || (bounds == null)) {
                            continue;
                        }
                        hasChunkLines = true;
                        runWords.AddChunk(text, bounds);
                    }
                }
            }
            if ((hasChunkLines) && (Model.AttachedList != null)) {
                var list = Model.AttachedList;
                runWords.AddListItem(list.ListLevels.GetListStyle(Model.AttachedListLevel), list.GenerateLevelText(Model));
            }
            foreach (var run in Model) {
                runWords.AddRun(run);
            }

            foreach (var obj in layout) {
                var layoutLine = obj as LayoutParagraphLine;
                if (layoutLine == null) {
                    continue;
                }
                foreach (var obj2 in layoutLine) {
                    var layoutWord = obj2 as LayoutParagraphWord;
                    if (layoutWord == null) {
                        continue;
                    }
                    runWords.AddWord(layoutWord);
                }
            }
            
            int pageNumber = Page.Index + Page.HeaderFooterInfo.PageNumberOffset - 1;
            var rws = runWords.Calculate(Page.HeaderFooterInfo.PageNumberFormat.FormatPageNumber(pageNumber));
            Hyperlink hyperlink = null;
            int i = 0;
            while (i < rws.Count) {
                var rw = rws[i];
                var textPosition = new PointF(rw.Word.Bounds.Left, rw.Word.Bounds.Top);
                float height = rw.HeightWithDescender/*.Word.Size.Height*/, width = rw.Word.Size.Width;
                if ((TextRotatation > 85) && (TextRotatation < 95)) {
                    textPosition.X = (float)(x + rw.Word.Bounds.Top - y);
                    textPosition.Y = (float)(y - (width + rw.Word.Bounds.Left - x));
                    float temp = height;
                    height = width;
                    width = temp;
                } else if ((TextRotatation > 265) && (TextRotatation < 275)) {
                    textPosition.X = (float)(x + y - rw.Word.Bounds.Top - height);
                    textPosition.Y = (float)(y + rw.Word.Bounds.Left - x);
                    float temp = height;
                    height = width;
                    width = temp;
                }
                var s = new Dictionary<string, object> {
                    { "width", Pt(width) },
                    { "height", Pt(height) },
                    { "top", Pt(textPosition.Y-y) },
                    { "left", Pt(textPosition.X-x) }
                };

                var spans = new List<StyleSpan>();
                while ((i < rws.Count) && (rws[i].Word == rw.Word)) {
                    string text = rws[i].Text/*.Trim()*/.Replace(char.ToString((char) 14), "");
                    var run = rws[i].RunInfo.Run;
                    var hyperlink2 = run != null? run.Hyperlink : null;
                    var bookmark = run != null? run.BookmarkStart : null;
                    spans.Add(new StyleSpan(this, rws[i].RunInfo.Style, text, hyperlink2, bookmark));
                    i += 1;
                }
                bool trickyHyperlinkChange = false;
                int j = 1;
                while (j < spans.Count) {
                    if (spans[j].CanMergeWith(spans[j - 1])) {
                        spans[j-1].Text += spans[j].Text;
                        spans.RemoveAt(j);
                        continue;
                    } else if (spans[j - 1].Bookmark != null && string.IsNullOrEmpty(spans[j - 1].Text) && spans[j].Bookmark == null) {
                        spans[j].Bookmark = spans[j - 1].Bookmark;
                        spans[j - 1] = spans[j];
                        spans.RemoveAt(j);
                        continue;
                    }
                    if (spans[j].Hyperlink != spans[j - 1].Hyperlink) {
                        trickyHyperlinkChange = true;
                    }
                    j += 1;
                }
                if (spans.Count > 1) {
                    string backgroundColor = spans[0].BackgroundColor;
                    bool consistent = true;
                    foreach (var span in spans) {
                        if (backgroundColor != span.BackgroundColor) {
                            consistent = false;
                        }
                    }
                    if (consistent && backgroundColor != null) {
                        s["background"] = backgroundColor;
                        foreach (var span in spans) {
                            span.BackgroundColor = null;
                        }
                    }
                    if (trickyHyperlinkChange) {
                        DoHyperlink(html, null, hyperlink);
                    } else if (spans[0].Hyperlink != hyperlink) {
                        DoHyperlink(html, spans[0].Hyperlink, hyperlink);
                        hyperlink = spans[0].Hyperlink;
                    }
                    html.Open("span", "position", s);
                    if (trickyHyperlinkChange) {
                        DoHyperlink(html, hyperlink, null);
                    }
                    foreach (var span in spans) {
                        if (trickyHyperlinkChange && span.Hyperlink != hyperlink) {
                            DoHyperlink(html, span.Hyperlink, hyperlink);
                            hyperlink = span.Hyperlink;
                        }
                        s = new Dictionary<string, object>();
                        span.UpdateStyle(s, paragraphStyle);
                        Dictionary<string, object> a = null;
                        if (span.Bookmark != null) {
                            a = new Dictionary<string, object> { { "id", GetBookmarkId(span.Bookmark) } };
                        }
                        html.Inline("span", "style", span.Text, s, a);
                    }
                    if (trickyHyperlinkChange) {
                        DoHyperlink(html, null, hyperlink);
                    }
                    html.Close("span");
                    if (trickyHyperlinkChange) {
                        DoHyperlink(html, hyperlink, null);
                    }
                } else if (spans.Count > 0) {
                    if (spans[0].Hyperlink != hyperlink) {
                        DoHyperlink(html, spans[0].Hyperlink, hyperlink);
                        hyperlink = spans[0].Hyperlink;
                    }
                    spans[0].UpdateStyle(s, paragraphStyle);
                    Dictionary<string, object> a = null;
                    if (spans[0].Bookmark != null) {
                        a = new Dictionary<string, object> { { "id", GetBookmarkId(spans[0].Bookmark) } };
                    }
                    html.Inline("span", "position style", spans[0].Text, s, a);
                }
            }
            DoHyperlink(html, null, hyperlink);
            return true;
        }
        
        public void GetParagraphStyles(Dictionary<string, object> s) {
            var fonts = new List<string>();
            var sizes = new List<string>();
            var colors = new List<string>();
            var weights = new List<string>();
            //var highlights = new List<string>();
            foreach (var run in Model) {
                var style = run.Style;
                string fontname = (style.Font != null ? style.Font.Name : "arial");
                if (fontname.Contains(" ")) {
                    fontname = "'" + fontname + "'";
                }
                fonts.Add(fontname + ", serif");
                string fontSize = Pt(style.FontSize);
                if (style.FontSize < 0.25) {
                    fontSize = Pt(Math.Max(6, -style.FontSize));
                }
                sizes.Add(fontSize);
                colors.Add(style.HasTextColor? style.TextColor.ToHexString() : null);
                weights.Add(style.FontWeight != 0? style.FontWeight.ToString() : null);
                //highlights.Add(style.HasHighlightColor? style.HighlightColor.ToHexString() : null);
            }
            string best = GetMedian(fonts);
            if (best != null) {
                s["font-family"] = best;
            }
            if ((best = GetMedian(sizes)) != null) {
                s["font-size"] = best;
            }
            if ((best = GetMedian(colors)) != null) {
                s["color"] = best;
            }
            if ((best = GetMedian(weights)) != null) {
                s["font-weight"] = best;
            }
            /*if ((best = GetMedian(highlights)) != null) {
                s["background"] = best;
            }*/
        }

        private string GetMedian(List<string> options) {
            int max = 0;
            string best = null;
            while (options.Count > 0) {
                string next = options[0];
                int count = options.RemoveAll(x => x == next);
                if (count > max) {
                    max = count;
                    best = next;
                }
            }
            return best;
        }
    }

    class StyleSpan {
        private readonly string[] inheritableStyles = { "font-family", "font-size", "color", "font-weight" };
        private string fontName;
        private string fontSize;
        private string fontColor;
        private string fontWeight;
        private string fontStyle;
        private string fontDecoration;
        private string fontHighlight;
        private string fontStretch;
        public string Text;
        public Hyperlink Hyperlink;
        public Bookmark Bookmark;
        public String BackgroundColor { get { return fontHighlight; } set { fontHighlight = value; } }

        public StyleSpan(BaseModel parent, TextStyle style, string text, Hyperlink hyperlink, Bookmark bookmark) {
            Text = text;
            Hyperlink = hyperlink;
            Bookmark = bookmark;
            fontName = (style.Font != null ? style.Font.Name : "arial");
            if (fontName.Contains(" ")) {
                fontName = "'" + fontName + "'";
            }
            fontName += ", serif";
            fontSize = parent.Pt(style.FontSize);
            if (style.FontSize < 0.25) {
                fontSize = parent.Pt(Math.Max(6, -style.FontSize));
            }
            if (style.HasTextColor) {
                fontColor = style.TextColor.ToHexString();
            }
            if (style.Italic) {
                fontStyle = "italic";
            }
            if (style.FontWeight != 0) {
                fontWeight = style.FontWeight.ToString();
            }
            if (style.UnderlineType != BorderLineType.None) {
                fontDecoration = "underline";
            }
            if (style.HasHighlightColor) {
                fontHighlight = style.HighlightColor.ToHexString();
            }
            if ((style.TextScale > 0.2) && ((style.TextScale < 0.99) || (style.TextScale > 1.01))) {
                fontStretch = style.TextScale.ToString("0.00");
            }
        }

        public bool CanMergeWith(StyleSpan o) {
            if (o == null) { return false; }
            if (Hyperlink != o.Hyperlink) { return false; }
            if (fontName != o.fontName) { return false; }
            if (fontSize != o.fontSize) { return false; }
            if (fontColor != o.fontColor) { return false; }
            if (fontWeight != o.fontWeight) { return false; }
            if (fontStyle != o.fontStyle) { return false; }
            if (fontDecoration != o.fontDecoration) { return false; }
            if (fontHighlight != o.fontHighlight) { return false; }
            if (fontStretch != o.fontStretch) { return false; }
            return true;
        }

        public void UpdateStyle(Dictionary<string, object> styles, Dictionary<string, object> currentStyle) {
            if (fontName != null) {
                styles["font-family"] = fontName;
            }
            if (fontSize != null) {
                styles["font-size"] = fontSize;
            }
            if (fontColor != null) {
                styles["color"] = fontColor;
            }
            if (fontStyle != null) {
                styles["font-style"] = fontStyle;
            }
            if (fontWeight != null) {
                styles["font-weight"] = fontWeight;
            }
            if (fontDecoration != null) {
                styles["text-decoration"] = fontDecoration;
            }
            if (fontHighlight != null) {
                styles["background"] = fontHighlight;
            }
            if (fontStretch != null) {
                styles["transform"] = "ScaleX(" + fontStretch + ")";
            }
            foreach (string key in inheritableStyles) {
                if (styles.ContainsKey(key) && currentStyle.ContainsKey(key)) {
                    if (styles[key].ToString() == currentStyle[key].ToString()) {
                        styles.Remove(key);
                    }
                }
            }
        }
    }
}
