﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing.Imaging;
using System.IO;
using SolidFramework.Model.Plumbing;
using TextBox = SolidFramework.Model.Plumbing.TextBox;

namespace PdfToHtml.Models { 

    class TextBoxModel: ParentModel {
        public TextBox Model { get; private set; }

        public TextBoxModel(ParentModel parent, TextBox textBox) : base(parent) {
            ZIndex = Parent.ZIndex;
            Model = textBox;
            ZIndex = Model.GetZIndex(ZIndex);
            
            if (!GetLayoutInfo(Model)) {
                Debug.Print("TextBox doesn't have a LayoutObject");
                Left = Model.Transform.XOffset;
                Top = Model.Transform.YOffset;
                Width = Model.Width;
                Height = Model.Height;
                var group = Parent as ShapeGroupModel;
                if (group != null) {
                    Left += group.Left;
                    Top += group.Top;
                }
            }
            

            int rotation = Model.Rotation;
            if (rotation == 0) {
                var textDirection = Model.TextDirection;
                if (textDirection == TextDirection.Rotate90) {
                    rotation = 90;
                } else if (textDirection == TextDirection.Rotate270) {
                    rotation = 270;
                }
            }
            TextRotatation = (TextRotatation + rotation) % 360;
            foreach (var obj in Model) {
                AddContent(obj);
            }
        }

        public override void SaveHtml(HtmlBuilder html, double x, double y) {
            var fill = Model.Fill as SolidFill;
            var line = Model.Line;
            var s = new Dictionary<string, object> {
                { "width", Pt(Width) },
                { "height", Pt(Height) },
                { "display", "block" },
                { "z-index", ZIndex },
                { "top", Pt(Top-y) },
                { "left", Pt(Left-x) }
            };
            if (fill != null) {
                s["background"] = fill.Color.ToHexString();
            }
            if (line != null) {
                s["border"] = Pt(line.Weight) + " solid " + line.Color.ToHexString();
            }
            html.Open("div", "textbox", s);
            if (TextRotatation > 10 || TextRotatation < -10) {
                html.Open("div", "rotation", new Dictionary<string, object> {
                    { "transform", "rotate(" + ((TextRotatation + 360) % 360) + "deg)" }
                });
            }
            foreach (var child in Contents) {
                child.SaveHtml(html, Left, Top);
            }
            if (TextRotatation > 10 || TextRotatation < -10) {
                html.Close("div");
            }
            html.Close("div");
        }
    }

    sealed class ImageModel: BaseModel {
        public static int NextImageNumber;
        public int ImageNumber;
        public ImageShape Model { get; private set; }

        public ImageModel(ParentModel parent, ImageShape image) : base(parent) {
            ImageNumber = NextImageNumber++;
            ZIndex = Parent.ZIndex;
            Model = image;

            ZIndex = Model.GetZIndex(ZIndex);

            if (!GetLayoutInfo(Model)) {
                Debug.Print("ImageShape doesn't have a LayoutObject");
                Left = Model.Transform.XOffset;
                Top = Model.Transform.YOffset;
                Width = Model.Width;
                Height = Model.Height;
                var group = Parent as ShapeGroupModel;
                if (group != null) {
                    Left += group.Left;
                    Top += group.Top;
                }
            }
        }

        public override void SaveHtml(HtmlBuilder html, double x, double y) {
            var image = Model.GetImage().ToBitmap();
            string src;
            if (File.EmbedImages) {
                using (var stream = new MemoryStream()) {
                    image.Save(stream, ImageFormat.Png);
                    src = "data:image/png;base64," + Convert.ToBase64String(stream.ToArray());
                }
            } else {
                string path = html.AssetPath(ImageNumber + ".png");
                image.Save(path, ImageFormat.Png);
                src = html.AssetUrl(ImageNumber + ".png");
            }
            var s = new Dictionary<string, object> {
                { "width", Pt(Width) },
                { "height", Pt(Height) },
                { "display", "block" },
                { "z-index", ZIndex },
                { "top", Pt(Top-y) },
                { "left", Pt(Left-x) }
            };
            html.Inline("img", "image", null, s, new Dictionary<string, object> {
                { "src", src }
            });
        }
    }


    sealed class GraphicModel: BaseModel {
        public Graphic Model { get; private set; }

        public GraphicModel(ParentModel parent, Graphic graphic) : base(parent) {
            ZIndex = Parent.ZIndex;
            Model = graphic;

            ZIndex = Model.GetZIndex(ZIndex);

            if (!GetLayoutInfo(Model)) {
                try {
                    Left = Model.Transform.XOffset;
                    Top = Model.Transform.YOffset;
                    Width = Model.Width;
                    Height = Model.Height;
                } catch (Exception e) {
                    Left = 30;
                    Top = -30;
                    Width = 30;
                    Height = 30;
                    Debug.WriteLine("Exception getting Graphic bounds: " + e.Message);
                }
                var group = Parent as ShapeGroupModel;
                if (group != null) {
                    Left += group.Left;
                    Top += group.Top;
                }
            }

            if (Width < 2) {
                Width = 2;
            }
            if (Height < 2) {
                Height = 2;
            }
        }

        public override void SaveHtml(HtmlBuilder html, double x, double y) {
            var fill = Model.Fill as SolidFill;
            //todo support non solid fills
            var line = Model.Line;
            double strokeWeight = line != null? line.Weight : 0;
            double minWidth = Model.Width;
            double minHeight = Model.Height;
            double offsetX = 0, offsetY = 0;
            if (minWidth < 0.25) {
                offsetX = -strokeWeight / 2;
                minWidth = strokeWeight;
            }
            if (minHeight < 0.25) {
                offsetY = -strokeWeight / 2;
                minHeight = strokeWeight;
            }
            if (minHeight < 1) {
                minHeight = 1;
            }
            if (minWidth < 1) {
                minWidth = 1;
            }
            var s = new Dictionary<string, object> {
                { "width", Pt(minWidth) },
                { "height", Pt(minHeight) },
                { "display", "block" },
                { "z-index", ZIndex },
                { "top", Pt(Top-y+offsetY) },
                { "left", Pt(Left-x+offsetX) }
            };
            html.Open("svg", "graphic", s, new Dictionary<string, object> {
                { "viewbox", offsetX + ", " + offsetY + ", " + minWidth + ", " + minHeight }
            });
            
            foreach (var nativePath in Model) {
                var segments = nativePath.GetSegments();
                if ((segments == null) || (segments.Count < 1)) {
                    continue;
                }

                var a = new Dictionary<string, object>();
                string definition = "";
                bool first = true;
                foreach (var segment in segments) {
                    if (first) {
                        first = false;
                        definition += "M " + segment.Start.X + " " + segment.Start.Y;
                    }
                    var bez = segment as BezierGraphicSegment;
                    if (bez != null) {
                        definition += " C " + bez.StartHandle.X + " " + bez.StartHandle.Y;
                        definition += ", " + bez.EndHandle.X + " " + bez.EndHandle.Y;
                        definition += ", " + bez.End.X + " " + bez.End.Y;
                    } else {
                        definition += " L " + segment.End.X + " " + segment.End.Y;
                    }
                }
                if (nativePath.IsClosed) {
                    definition += " Z";
                }

                a["d"] = definition;
                a["fill"] = fill != null? fill.Color.ToHexString() : "none";
                if (fill != null) {
                    a["fill-opacity"] = fill.Color.A > 0 ? fill.Color.A / 255.0 : 1;
                }
                a["stroke"] = line != null? line.Color.ToHexString() : "none";
                if (line != null) {
                    a["stroke-opacity"] = line.Color.A > 0 ? line.Color.A / 255.0 : 1;
                    a["stroke-width"] = strokeWeight;
                    if (line.DashType != LineDashType.Solid) {
                        a["stroke-dasharray"] = "3, 4, 3, 4";
                        //TODO support dashes properly
                    }
                }
                html.Inline("path", attributes: a);
            }
            html.Close("svg");
        }
    }


    sealed class ShapeGroupModel: ParentModel {
        public Group Model { get; private set; }

        public ShapeGroupModel(ParentModel parent, Group group) : base(parent) {
            ZIndex = Parent.ZIndex;
            Model = group;

            ZIndex = Model.GetZIndex(ZIndex);
            
            if (!GetLayoutInfo(Model)) {
                Debug.Print("Group doesn't have a LayoutObject");
                Left = Model.Transform.XOffset;
                Top = Model.Transform.YOffset;
                Width = Model.Width;
                Height = Model.Height;
            }

            foreach (var obj in Model) {
                if ((obj is TextBox) || (obj is ImageShape) || (obj is Group) || (obj is Graphic)) {
                    AddContent(obj);
                } else {
                    throw new InvalidOperationException("Error: Unexpected object type '" + obj.GetType() + "' found inside Group");
                }
            }
        }

        public override void SaveHtml(HtmlBuilder html, double x, double y) {
            var fill = Model.Fill as SolidFill;
            var line = Model.Line;
            var s = new Dictionary<string, object> {
                { "width", Pt(Width) },
                { "height", Pt(Height) },
                { "display", "block" },
                { "z-index", ZIndex },
                { "top", Pt(Top-y) },
                { "left", Pt(Left-x) }
            };
            if (fill != null) {
                s["background"] = fill.Color.ToHexString();
            }
            if (line != null) {
                s["border"] = Pt(line.Weight) + " solid " + line.Color.ToHexString();
            }
            html.Open("div", "group", s);
            foreach (var child in Contents) {
                child.SaveHtml(html, Left, Top);
            }
            html.Close("div");
        }
    }
}
