﻿using SolidFramework.Model.Plumbing;
using System;
using System.Collections.Generic;
using System.Linq;

namespace PdfToHtml.Models {
    sealed class TableModel: ParentModel {
        public Table Model { get; private set; }
        
        public List<RowModel> Rows { get; private set; }

        public double CellSpacing { get; private set; }

        public Dictionary<int, CellModel> CellMap { get; private set; }
        
        public List<double> ColumnPositions;

        public TableModel(ParentModel parent, Table table): base(parent) {
            ZIndex = Parent.ZIndex;
            Model = table;
            Rows = new List<RowModel>();
            CellMap = new Dictionary<int, CellModel>();


            if (!GetLayoutInfo(Model)) {
                Left = table.Transform.XOffset;
                Top = table.Transform.YOffset;
            }
            Width = 0;
            Height = 0;

            CellSpacing = Model.CellSpacing;
            
            foreach (var obj in table.Rows) {
                var row = new RowModel(this, obj);
                Rows.Add(row);
                Contents.Add(row);
                Height += row.Height + CellSpacing;
            }

            var columnPositions = new Dictionary<int, double>();

            foreach (var row in Rows) {
                Width = Math.Max(row.Width, Width);
                double left = 0;
                foreach (var cell in row.Cells) {
                    left += cell.Width;
                    columnPositions[(int)((left + 2) / 4)] = left;
                }
            }
            ColumnPositions = columnPositions.Values.ToList();
            ColumnPositions.Sort();
            foreach (var row in Rows) {
                double left = 0;
                int col = 0;
                foreach (var cell in row.Cells) {
                    left += cell.Width;
                    cell.ColSpan = 0;
                    while (col + cell.ColSpan < ColumnPositions.Count && ColumnPositions[col + cell.ColSpan] <= left) {
                        cell.ColSpan += 1;
                    }
                    col += cell.ColSpan;
                }
            }
        }

        public override void SaveHtml(HtmlBuilder html, double x, double y) {
            html.Open("table", "table", new Dictionary<string, object> {
                { "width", Pt(Width) },
                { "height", Pt(Height) },
                { "position", "absolute" },
                { "z-index", ZIndex },
                { "top", Pt(Top-y)},
                { "left", Pt(Left-x) },
                { "border-collapse", "collapse" }
            });
            html.Open("tbody");
            html.Open("tr");
            for (int i = 0; i < ColumnPositions.Count; i++) {
                html.Inline("td", style: new Dictionary<string, object> {
                    { "width", Pt(ColumnPositions[i] - (i == 0? 0 : ColumnPositions[i - 1]), "0.00") },
                    { "border", "none" },
                    { "height", Pt(0, "0") }
                });
            }
            html.Close("tr");
            foreach (var row in Rows) {
                html.Open("tr");
                foreach (var cell in row.Contents) {
                    cell.SaveHtml(html, Left, Top);
                }
                html.Close("tr");
            }
            html.Close("tbody");
            html.Close("table");
        }
    }

    sealed class RowModel: ParentModel {
        public TableModel Table{ get { return (TableModel)Parent;}}
        public Row Model { get; private set; }
        public int RowNumber { get; private set; }

        public List<CellModel> Cells { get; private set; }
        
        public double TopBorderSize { get; private set; }
        public double BottomBorderSize { get; private set; }

        private void GetBorderSize() {
            var cellsFromAbove = new HashSet<int>();
            var cellsFromBelow = new HashSet<int>();
            var rowList = Table.Model.Rows;
            if (Index > 1) {
                foreach (var cell in rowList[Index-2]) {
                    cellsFromAbove.Add(cell.GetID());
                }
            }
            if (rowList.Count > Index) {
                foreach (var cell in rowList[Index]) {
                    cellsFromBelow.Add(cell.GetID());
                }
            }
            TopBorderSize = 0;
            BottomBorderSize = 0;
            foreach (var cell in Model) {
                int id = cell.GetID();
                var borders = cell.Borders;
                if (!cellsFromAbove.Contains(id)) {
                    TopBorderSize = Math.Max(borders.Top.GetBorderSize(), TopBorderSize);
                }
                if (!cellsFromBelow.Contains(id)) {
                    BottomBorderSize = Math.Max(borders.Bottom.GetBorderSize(), BottomBorderSize);
                }
            }
        }

        private double CalculateRowHeight() {
            double height = Model.Height;
            if (Model.GetType().GetMethod("CalculateWordHeight") != null) {
                return height; // builds that include a function to calculate the word height will have absolute row heights
            }
            // Rows in Document mode often/always? have a RowHeightRule set to 'AtLeast' which means their height won't include the size of the border
            height += TopBorderSize / 2;
            height += BottomBorderSize / 2;
            return height;
        }

        public RowModel(TableModel table, Row row): base(table) {
            ZIndex = Parent.ZIndex;
            Model = row;
            RowNumber = Table.Rows.Count;
            Cells = new List<CellModel>();

            GetBorderSize();
            if (Index == 1) {
                Table.Top += TopBorderSize / 2;
            }

            Left = Table.Left;
            Top = Table.Top + Table.Height;
            Width = 0;
            Height = CalculateRowHeight();

            if (Index == Table.Model.Rows.Count) {
                Table.Height += BottomBorderSize / 2;
            }

            foreach (var cell in Model) {
                if (Table.CellMap.ContainsKey(cell.GetID())) {
                    var existingCell = Table.CellMap[cell.GetID()];
                    existingCell.UpdateRows(this);
                    Cells.Add(existingCell);
                    Width += existingCell.Width + Table.CellSpacing;
                    continue;
                }
                var Cell = new CellModel(this, cell);
                Table.CellMap.Add(cell.GetID(), Cell);
                Cells.Add(Cell);
                Contents.Add(Cell);
                Width += Cell.Width + Table.CellSpacing;
            }
        }
    }

    sealed class CellModel: ParentModel {
        public RowModel Row { get { return (RowModel)Parent; } }
        public Cell Model { get; private set; }

        public int ColSpan { get; set; }
        public int RowSpan { get; set;  }

        public TextDirection TextDirection { get; private set; }

        public CellModel(RowModel row, Cell cell): base(row) {
            ZIndex = Parent.ZIndex;
            Model = cell;

            int columnNumber = 0;
            foreach (var other in Row.Cells) {
                columnNumber += other.ColSpan;
            }

            ColSpan = 1;
            var nextCell = Row.Table.Model.GetCell(Row.RowNumber, columnNumber + ColSpan);
            while ((nextCell != null) && (nextCell.GetID() == cell.GetID())) {
                ColSpan += 1;
                nextCell = Row.Table.Model.GetCell(Row.RowNumber, columnNumber + ColSpan);
            }

            Left = Row.Left + Row.Width;
            Top = Row.Top;
            Width = cell.Width;
            Height = Row.Height;

            TextDirection = Model.TextDirection;
            int rotation = 0;
            if (TextDirection == TextDirection.Rotate90) {
                rotation = 90;
            } else if (TextDirection == TextDirection.Rotate270) {
                rotation = 270;
            }
            TextRotatation = (TextRotatation + rotation) % 360;

            foreach (var obj in Model) {
                AddContent(obj);
            }
            RowSpan = 0;
            UpdateRows(Row);
        }

        public override void SaveHtml(HtmlBuilder html, double x, double y) {
            var s = new Dictionary<string, object> {
                { "width", Pt(Width, "0.00") },
                { "height", Pt(Height, "0.00") } // inaccuracy in a table can really build up
            };
            if (Model.HasFillColor) {
                s["background"] = Model.FillColor.ToHexString();
            }
            
            var borders = Model.Borders;
            AddBorderCSS(s, borders.Left, "Left");
            AddBorderCSS(s, borders.Top, "Top");
            AddBorderCSS(s, borders.Right, "Right");
            AddBorderCSS(s, borders.Bottom, "Bottom");
            html.Open("td", "cell", s, new Dictionary<string, object> {
                { "rowspan", RowSpan },
                { "colspan", ColSpan }
            });
            //todo support rotation
            foreach (var child in Contents) {
                child.SaveHtml(html, x, y);
            }
            html.Close("td");
        }

        private void AddBorderCSS(Dictionary<string, object> s, LineStyle border, string side) {
            if (border == null) {
                return;
            }
            s["border-" + side.ToLower()] = Pt(border.GetBorderSize()) + " solid " + border.Color.ToHexString();
        }
        
        internal void UpdateRows(RowModel row) {
            RowSpan += 1;
            if (row != Row) {
                Height += row.Height + Row.Table.CellSpacing;
            }
        }
    }
}
