﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using SolidFramework.Model.Layout;
using SolidFramework.Model.Plumbing;

namespace PdfToHtml {

    public class RunInfo {
        public Run Run { get; set; } // will be null for a list
        public TextStyle Style { get; set; }
        public bool IsPageNumberField { get; set; }
    }
    
    public class WordLayout {
        public RectangleF Bounds { get; set; }
        public SizeF Size { get; set; }
        public string Text { get; set; }

        public static implicit operator WordLayout(LayoutParagraphWord layoutWord) {
            return new WordLayout {
                Bounds = layoutWord.Bounds,
                Size = layoutWord.Size,
                Text = layoutWord.Text
            };
        }
    }

    public class ChunkLayout: WordLayout {
        public ChunkLayout(string text, RectangleF bounds) {
            Text = text;
            Bounds = bounds;
            Size = Bounds.Size;
        }
    }

    public class RunWord {
        public RunInfo RunInfo { get; set; }
        public WordLayout Word { get; set; }
        public string Text { get; set; }
        public int WordOffset { get; set; }
        public int RunOffset { get; set; }
        
        public float HeightWithDescender {
            get {
                /*if (RunInfo.Style != null && RunInfo.Style.Font != null) {
                    var font = new FontFamily(RunInfo.Style.Font.Name);
                    var fontStyle = FontStyle.Regular;
                    if (RunInfo.Style.Italic) {
                        fontStyle &= FontStyle.Italic;
                    }
                    if (RunInfo.Style.FontWeight > 550) {
                        fontStyle &= FontStyle.Bold;
                    }
                    float descendentHeight = font.GetCellDescent(fontStyle);
                    float emHeight = font.GetEmHeight(fontStyle);
                    return Word.Size.Height + Word.Size.Height * descendentHeight / emHeight;
                }*/ 
                return Word.Size.Height;
            }
        }
    }

    public class RunWords {
        private readonly List<WordLayout> chunks;
        private readonly List<WordLayout> words;
        private readonly List<RunInfo> runs;
        private readonly List<string> chunksText;
        private readonly List<string> wordsText;
        private readonly List<string> runsText;

        public RunWords() {
            chunks = new List<WordLayout>();
            words = new List<WordLayout>();
            runs = new List<RunInfo>();
            chunksText = new List<string>();
            wordsText = new List<string>();
            runsText = new List<string>();
        }

        public void AddChunk(string text, RectangleF bounds) {
            chunks.Add(new ChunkLayout(text, bounds));
            chunksText.Add(text);
        }

        public void AddWord(WordLayout word) {
            words.Add(word);
            wordsText.Add(word.Text + " ");
        }

        public void AddRun(Run run) {
            runs.Add(new RunInfo {
                Run = run,
                Style = run.Style,
                IsPageNumberField = run.IsPageNumberField
            });
            runsText.Add(run.Text);
        }
        public void AddListItem(ListStyle style, string text) {
            runs.Add(new RunInfo {
                Run = null,
                Style = style,
                IsPageNumberField = false
            });
            runsText.Add(text);
        }

        public List<RunWord> Calculate(string pageNumber="{page-number}") {
            var wordList = words;
            var wordTexts = wordsText;
            if (chunks.Count > 0) {
                wordList = chunks;
                wordTexts = chunksText;
            }
            if ((runs.Count < 1) || (wordList.Count < 1)) {
                return new List<RunWord>();
            }
            string wordText = "", runText = "";
            foreach (string text in wordTexts) {
                wordText += text;
            }
            foreach (string text in runsText) {
                runText += text.Replace("\t", " ");
            }
            var changes = MyersDiff.Diff(runText, wordText);
            changes.Reverse();
            var state = new List<byte>();
            int i;
            for (i = 0; i < runText.Length; i++) {
                state.Add(0);
            }
            foreach (var change in changes) {
                if (change.Type == DiffOperationType.Delete) {
                    //state.Insert(change.Index, 2);
                    state[change.Index] = 2;
                } else {
                    state.Insert(change.Index, 1);
                }
            }
            var runWords = new List<RunWord>();
            int wordIndex = 0, wordOffset = 0;
            int runIndex = 0, runOffset = 0, prevRunOffset = 0;
            while ((wordIndex < wordTexts.Count) && (wordTexts[wordIndex].Length < 1)) {
                wordIndex += 1;
            }
            while ((runIndex < runsText.Count) && (runsText[runIndex].Length < 1)) {
                runWords.Add(new RunWord {
                    RunInfo = runs[runIndex],
                    Word = wordList.Count > 0? wordList[wordIndex < wordList.Count? wordIndex : wordIndex - 1] : null,
                    Text = "",
                    RunOffset = runOffset,
                    WordOffset = 0
                });
                runIndex += 1;
            }
            for (i = 0; i < state.Count; i++) {
                if ((wordIndex >= wordTexts.Count) || (runIndex >= runsText.Count)) {
                    break;
                }
                if (state[i] == 0) {
                    wordOffset += 1;
                    runOffset += 1;
                } else if (state[i] == 2) {
                    runOffset += 1;
                } else {
                    wordOffset += 1;
                }
                if ((wordOffset < wordTexts[wordIndex].Length) && (runOffset < runsText[runIndex].Length)) {
                    continue;
                }
                runWords.Add(new RunWord {
                    RunInfo = runs[runIndex],
                    Word = wordList[wordIndex],
                    Text = runsText[runIndex].Substring(prevRunOffset, runOffset - prevRunOffset),
                    RunOffset = runOffset,
                    WordOffset = wordOffset
                });
                if (wordOffset >= wordTexts[wordIndex].Length) {
                    wordIndex += 1;
                    wordOffset = 0;
                }
                if (runOffset >= runsText[runIndex].Length) {
                    runIndex += 1;
                    runOffset = 0;
                }
                prevRunOffset = runOffset;

                while ((wordIndex < wordTexts.Count) && (wordTexts[wordIndex].Length < 1)) {
                    wordIndex += 1;
                }
                while ((runIndex < runsText.Count) && (runsText[runIndex].Length < 1)) {
                    runWords.Add(new RunWord {
                        RunInfo = runs[runIndex],
                        Word = wordList.Count > 0? wordList[wordIndex < wordList.Count? wordIndex : wordIndex - 1] : null,
                        Text = "",
                        RunOffset = runOffset,
                        WordOffset = 0
                    });
                    runIndex += 1;
                }
            }
            var lastRun = runs.LastOrDefault();
            var lastWord = wordList.LastOrDefault();
            while (i < state.Count) {
                if (runIndex < runs.Count) {
                    var runWord = new RunWord {
                        RunInfo = runs[runIndex],
                        Word = lastWord,
                        Text = runsText[runIndex].Substring(runOffset),
                        RunOffset = runOffset,
                        WordOffset = 0
                    };
                    runWords.Add(runWord);
                    runOffset = 0;
                    runIndex += 1;
                } else if (wordIndex < wordList.Count) {
                    var runWord = new RunWord {
                        RunInfo = lastRun,
                        Word = wordList[wordIndex],
                        Text = "",
                        RunOffset = 0,
                        WordOffset = wordOffset
                    };
                    runWords.Add(runWord);
                    wordOffset = 0;
                    wordIndex += 1;
                }
                i += 1;
            }
            bool samePageNumber = false; // avoid adding multiple copies of the same page number due to the page number having multiple chunks
            foreach (var rw in runWords) {
                if (rw.RunInfo.IsPageNumberField) {
                    if (rw.Text.Length > 0 && !samePageNumber) {
                        samePageNumber = true;
                        rw.Text = pageNumber; // should probably actually look at wordOffset and do this properly
                    }
                } else {
                    samePageNumber = false;
                }
            }
            return runWords;
        }
    }

    public enum DiffOperationType {
        Delete,
        Insert
    }

    public class DiffOperation {
        public DiffOperationType Type { get; set; }
        public int Index { get; set; }
        public int SourceIndex { get; set; }
        public char Value { get; set; }

        public static DiffOperation Delete(int index, char value) {
            return new DiffOperation {
                Type = DiffOperationType.Delete, 
                Index = index, 
                Value = value
            };
        }

        public static DiffOperation Insert(int index, int sourceIndex, char value) {
            return new DiffOperation {
                Type = DiffOperationType.Insert, 
                Index = index, 
                SourceIndex = sourceIndex, 
                Value = value
            };
        }

        public override string ToString() {
            if (Type == DiffOperationType.Insert) {
                return "Insert " + Value + " @" + Index;
            } else {
                return "Delete " + Value + " @" + Index;
            }
        }
    }

    public static class MyersDiff {

        private static int Mod(int value, int n) {
            value = value % n;
            return value >= 0 ? value : value + n;
        }

        private static void ReverseIndices(ref int initial, ref int current, int max) {
            int temp = initial;
            initial = max - current;
            current = max - temp;
        }
        
        public static List<DiffOperation> Diff(string a, string b, int offsetA=0, int offsetB=0) {
            int n = a.Length, m = b.Length;
            // handle empty cases
            if ((m < 1) || (n < 1)) {
                var ops = new List<DiffOperation>();
                for (int i = 0; i < n; i++) {
                    ops.Add(DiffOperation.Delete(offsetA + i, a[i]));
                }
                for (int i = 0; i < m; i++) {
                    ops.Add(DiffOperation.Insert(offsetA, offsetB + i, b[i]));
                }
                return ops;
            }
            // recursive diff
            int maxRange = m + n;
            int vSize = 2 * (Math.Min(m, n) + 1);
            var vForward = new int[vSize];
            var vReverse = new int[vSize];
            int oneIfOdd = maxRange % 2;
            int maxSearch = (maxRange / 2) + oneIfOdd;
            for (int currentSearch = 0; currentSearch <= maxSearch; currentSearch++) {
                for (int oneIfReverse = 0; oneIfReverse <= 1; oneIfReverse++) { // first walk from top-right, then from bottom-left
                    var v = (oneIfReverse == 0)? vForward : vReverse;
                    var vOther = (oneIfReverse == 0) ? vReverse : vForward;
                    for (int k = -(currentSearch - (2 * Math.Max(0, currentSearch - m))); k <= (currentSearch - (2 * Math.Max(0, currentSearch - n))); k += 2) {
                        // take the best x position for v[k] as either v[k-1] + 1 (insertion from a) or (v-1)[k] (deletion from a) 
                        int x = k == -currentSearch ? v[Mod(k + 1, vSize)] : v[Mod(k - 1, vSize)] + 1;
                        if (k != -currentSearch && k != currentSearch) {
                            x = v[Mod(k - 1, vSize)] + 1;
                            int xOther = v[Mod(k + 1, vSize)];
                            if (xOther >= x) {
                                x = xOther;
                            }
                        }
                        // increase x & y until a[x] != b[y]
                        int y = x - k; // k defines a line from (0, k) on the bottom-left to (k, 0) on the top-right, so y = x - k
                        int initialX = x, initialY = y;
                        while ((x < n) && (y < m) && (a[oneIfReverse == 0 ? x : n - x - 1] == b[oneIfReverse == 0 ? y : m - y - 1])) {
                            x += 1;
                            y += 1;
                        }
                        // store the x position of this chain
                        v[Mod(k, vSize)] = x;
                        // check if the forward and reverse search chains have met in the middle yet
                        int kOther = n - (m + k);
                        if ((oneIfOdd == oneIfReverse) || (Math.Abs(kOther) >= currentSearch + oneIfReverse) || (v[Mod(k, vSize)] + vOther[Mod(kOther, vSize)] < n)) {
                            continue;
                        }
                        if (oneIfReverse != 0) {
                            ReverseIndices(ref initialX, ref x, n);
                            ReverseIndices(ref initialY, ref y, m);
                        }
                        if ((2 * currentSearch + oneIfReverse > 2) || ((initialX != x) && (initialY != y))) {
                            var ops = new List<DiffOperation>();
                            ops.AddRange(Diff(a.Substring(0, initialX), b.Substring(0, initialY), offsetA, offsetB));
                            ops.AddRange(Diff(a.Substring(x), b.Substring(y), offsetA + x, offsetB + y));
                            return ops;
                        } else if (m > n) {
                            return Diff("", b.Substring(n), offsetA + n, offsetB + n);
                        } else if (n > m) {
                            return Diff(a.Substring(m), "", offsetA + m, offsetB + m);
                        } else {
                            return new List<DiffOperation>();
                        }
                    }
                }
            }
            return new List<DiffOperation>(); // shouldn't ever happen
        }
    }
}
